/**
 * Java Card implementation example for use with Radboud Reader
 * Copyright (C) 2013  Joeri de Ruiter
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package radboudcard;

import javacard.framework.*;
import javacard.security.*;
import javacardx.crypto.Cipher;
import Constants;

/**
 * @author Joeri de Ruiter (joeri@cs.ru.nl)
 */
public class RadboudCard extends Applet {
	byte[] tmp;
	
	byte[] state;
	byte[] ssc;
	
	DESKey 			sharedKey;
	RSAPublicKey	cardPublicKey;
	RSAPrivateKey	cardPrivateKey;
	RSAPublicKey	issuerPublicKey;

	DESKey			sessionKey;
	
	Cipher			rsaCipher;
	Cipher			desCipher;
	Signature		desMAC;
	
	RandomData		random;
	
	OwnerPIN		pin;
	
	byte	INT_STATE_INIT = 0x00;
	byte	INT_STATE_DISPLAY = 0x01;
	byte	INT_STATE_ASK_PIN = 0x02;
	
	short	_0 = 0;
	
	public RadboudCard() {
		tmp = JCSystem.makeTransientByteArray((short) 256, JCSystem.CLEAR_ON_DESELECT);
		state = JCSystem.makeTransientByteArray((short) 2, JCSystem.CLEAR_ON_DESELECT);
		ssc = JCSystem.makeTransientByteArray((short) 16, JCSystem.CLEAR_ON_DESELECT);
		
		sharedKey = (DESKey)KeyBuilder.buildKey(KeyBuilder.TYPE_DES, KeyBuilder.LENGTH_DES3_2KEY, false);
		sharedKey.setKey(new byte[]{(byte)0x00,(byte)0x01,(byte)0x02,(byte)0x03,(byte)0x04,(byte)0x05,(byte)0x06,(byte)0x07,(byte)0x08,(byte)0x09,(byte)0x0A,(byte)0x0B,(byte)0x0C,(byte)0x0D,(byte)0x0E,(byte)0x0F}, (short)0);
		
		sessionKey = (DESKey)KeyBuilder.buildKey(KeyBuilder.TYPE_DES_TRANSIENT_DESELECT, KeyBuilder.LENGTH_DES3_2KEY, false);
				
		cardPublicKey = (RSAPublicKey)KeyBuilder.buildKey(KeyBuilder.TYPE_RSA_PUBLIC, KeyBuilder.LENGTH_RSA_1024, false);
		cardPublicKey.setModulus(new byte[]{(byte)0x8B,(byte)0x2D,(byte)0xAD,(byte)0x98,(byte)0x3E,(byte)0x0B,(byte)0x14,(byte)0x0C,(byte)0x85,(byte)0x67,(byte)0x1B,(byte)0x06,(byte)0x6A,(byte)0xF6,(byte)0x22,(byte)0x90,(byte)0x45,(byte)0xB9,(byte)0x49,(byte)0x9B,(byte)0x56,(byte)0xDC,(byte)0xEF,(byte)0xB7,(byte)0xA1,(byte)0x3E,(byte)0xBB,(byte)0xC1,(byte)0x3E,(byte)0xDD,(byte)0x4D,(byte)0x0D,(byte)0x41,(byte)0xF2,(byte)0x58,(byte)0x57,(byte)0x7E,(byte)0x7D,(byte)0x54,(byte)0xA9,(byte)0xDC,(byte)0x30,(byte)0x7A,(byte)0xE7,(byte)0xF9,(byte)0x86,(byte)0x02,(byte)0x5D,(byte)0xB4,(byte)0x38,(byte)0x9E,(byte)0xD2,(byte)0x8F,(byte)0xB9,(byte)0x64,(byte)0xBF,(byte)0xEF,(byte)0x24,(byte)0xE0,(byte)0x9A,(byte)0xB0,(byte)0xEA,(byte)0x18,(byte)0x05,(byte)0x5C,(byte)0x26,(byte)0x4A,(byte)0x7F,(byte)0x6F,(byte)0x93,(byte)0x38,(byte)0x55,(byte)0xC7,(byte)0x51,(byte)0x54,(byte)0x50,(byte)0x72,(byte)0xD6,(byte)0x1D,(byte)0xCE,(byte)0x1D,(byte)0xBE,(byte)0x6F,(byte)0x63,(byte)0x25,(byte)0x90,(byte)0x1E,(byte)0x8E,(byte)0xD9,(byte)0xE1,(byte)0x36,(byte)0x61,(byte)0x49,(byte)0x01,(byte)0x20,(byte)0xF4,(byte)0xC9,(byte)0x71,(byte)0x7F,(byte)0x4F,(byte)0x5E,(byte)0x06,(byte)0xBB,(byte)0x7B,(byte)0x6C,(byte)0x6F,(byte)0x24,(byte)0x50,(byte)0xF9,(byte)0xD4,(byte)0xD2,(byte)0xF8,(byte)0x29,(byte)0xC1,(byte)0xDF,(byte)0x2A,(byte)0x4C,(byte)0x75,(byte)0x4F,(byte)0xC5,(byte)0xA8,(byte)0x9B,(byte)0x26,(byte)0x25,(byte)0xFE,(byte)0x1B,(byte)0x03,(byte)0x33}, (short)0, (short)128);
		cardPublicKey.setExponent(new byte[]{(byte)0x01,(byte)0x00,(byte)0x01}, (short)0, (short)3);

		cardPrivateKey = (RSAPrivateKey)KeyBuilder.buildKey(KeyBuilder.TYPE_RSA_PRIVATE, KeyBuilder.LENGTH_RSA_1024, false);
		cardPrivateKey.setModulus(new byte[]{(byte)0x8B,(byte)0x2D,(byte)0xAD,(byte)0x98,(byte)0x3E,(byte)0x0B,(byte)0x14,(byte)0x0C,(byte)0x85,(byte)0x67,(byte)0x1B,(byte)0x06,(byte)0x6A,(byte)0xF6,(byte)0x22,(byte)0x90,(byte)0x45,(byte)0xB9,(byte)0x49,(byte)0x9B,(byte)0x56,(byte)0xDC,(byte)0xEF,(byte)0xB7,(byte)0xA1,(byte)0x3E,(byte)0xBB,(byte)0xC1,(byte)0x3E,(byte)0xDD,(byte)0x4D,(byte)0x0D,(byte)0x41,(byte)0xF2,(byte)0x58,(byte)0x57,(byte)0x7E,(byte)0x7D,(byte)0x54,(byte)0xA9,(byte)0xDC,(byte)0x30,(byte)0x7A,(byte)0xE7,(byte)0xF9,(byte)0x86,(byte)0x02,(byte)0x5D,(byte)0xB4,(byte)0x38,(byte)0x9E,(byte)0xD2,(byte)0x8F,(byte)0xB9,(byte)0x64,(byte)0xBF,(byte)0xEF,(byte)0x24,(byte)0xE0,(byte)0x9A,(byte)0xB0,(byte)0xEA,(byte)0x18,(byte)0x05,(byte)0x5C,(byte)0x26,(byte)0x4A,(byte)0x7F,(byte)0x6F,(byte)0x93,(byte)0x38,(byte)0x55,(byte)0xC7,(byte)0x51,(byte)0x54,(byte)0x50,(byte)0x72,(byte)0xD6,(byte)0x1D,(byte)0xCE,(byte)0x1D,(byte)0xBE,(byte)0x6F,(byte)0x63,(byte)0x25,(byte)0x90,(byte)0x1E,(byte)0x8E,(byte)0xD9,(byte)0xE1,(byte)0x36,(byte)0x61,(byte)0x49,(byte)0x01,(byte)0x20,(byte)0xF4,(byte)0xC9,(byte)0x71,(byte)0x7F,(byte)0x4F,(byte)0x5E,(byte)0x06,(byte)0xBB,(byte)0x7B,(byte)0x6C,(byte)0x6F,(byte)0x24,(byte)0x50,(byte)0xF9,(byte)0xD4,(byte)0xD2,(byte)0xF8,(byte)0x29,(byte)0xC1,(byte)0xDF,(byte)0x2A,(byte)0x4C,(byte)0x75,(byte)0x4F,(byte)0xC5,(byte)0xA8,(byte)0x9B,(byte)0x26,(byte)0x25,(byte)0xFE,(byte)0x1B,(byte)0x03,(byte)0x33}, (short)0, (short)128);
		cardPrivateKey.setExponent(new byte[]{(byte)0x31,(byte)0x58,(byte)0x6C,(byte)0xA8,(byte)0x01,(byte)0xB7,(byte)0x63,(byte)0xD0,(byte)0x5C,(byte)0xAC,(byte)0x05,(byte)0xC4,(byte)0xEC,(byte)0xBF,(byte)0x7F,(byte)0xD9,(byte)0xAF,(byte)0x1E,(byte)0xD6,(byte)0x9E,(byte)0x20,(byte)0xDA,(byte)0x0C,(byte)0x94,(byte)0x0A,(byte)0xB3,(byte)0x77,(byte)0x56,(byte)0x38,(byte)0xB7,(byte)0xB5,(byte)0xCD,(byte)0xF1,(byte)0x87,(byte)0xBF,(byte)0x12,(byte)0xA6,(byte)0x62,(byte)0x41,(byte)0x56,(byte)0x48,(byte)0x1F,(byte)0x58,(byte)0x34,(byte)0x4F,(byte)0x1F,(byte)0xD8,(byte)0x4E,(byte)0xE8,(byte)0xCA,(byte)0xFA,(byte)0x64,(byte)0xA5,(byte)0x58,(byte)0x92,(byte)0x04,(byte)0x0B,(byte)0x02,(byte)0x04,(byte)0x30,(byte)0x54,(byte)0x71,(byte)0x6B,(byte)0x3A,(byte)0x54,(byte)0x43,(byte)0xF3,(byte)0x12,(byte)0x19,(byte)0xED,(byte)0x2E,(byte)0x47,(byte)0x5D,(byte)0x82,(byte)0x35,(byte)0x2B,(byte)0xB4,(byte)0xE4,(byte)0x21,(byte)0x18,(byte)0x4A,(byte)0x2E,(byte)0xCA,(byte)0x07,(byte)0x6F,(byte)0xC7,(byte)0x7C,(byte)0xB6,(byte)0x43,(byte)0xEF,(byte)0xEF,(byte)0x98,(byte)0x82,(byte)0xEF,(byte)0x7E,(byte)0x2C,(byte)0x64,(byte)0xEB,(byte)0xAC,(byte)0x9A,(byte)0x3B,(byte)0xB8,(byte)0xF6,(byte)0x58,(byte)0xDD,(byte)0x55,(byte)0x0F,(byte)0x8D,(byte)0xBF,(byte)0xD5,(byte)0x9D,(byte)0xE3,(byte)0x6B,(byte)0xAB,(byte)0x54,(byte)0x8E,(byte)0x3E,(byte)0x34,(byte)0xE4,(byte)0x6A,(byte)0xE5,(byte)0xD8,(byte)0x1B,(byte)0x45,(byte)0xAA,(byte)0x58,(byte)0xA1,(byte)0x41}, (short)0, (short)128);
		
		issuerPublicKey = (RSAPublicKey)KeyBuilder.buildKey(KeyBuilder.TYPE_RSA_PUBLIC, KeyBuilder.LENGTH_RSA_1024, false);
		issuerPublicKey.setModulus(new byte[]{(byte)0xA4,(byte)0x51,(byte)0x12,(byte)0x2E,(byte)0x9F,(byte)0xB5,(byte)0x06,(byte)0xE5,(byte)0xF4,(byte)0x8F,(byte)0x73,(byte)0x1E,(byte)0xD7,(byte)0x50,(byte)0x1A,(byte)0xCC,(byte)0x6A,(byte)0xD2,(byte)0xF4,(byte)0x11,(byte)0xDF,(byte)0xD4,(byte)0xA5,(byte)0x8B,(byte)0xC3,(byte)0x7B,(byte)0xA0,(byte)0x09,(byte)0xA5,(byte)0xDA,(byte)0x90,(byte)0x51,(byte)0xD2,(byte)0xC3,(byte)0x26,(byte)0x58,(byte)0x1D,(byte)0x60,(byte)0xAC,(byte)0xB6,(byte)0xBA,(byte)0xDA,(byte)0xFD,(byte)0x4A,(byte)0x40,(byte)0x03,(byte)0x3C,(byte)0x24,(byte)0x26,(byte)0x98,(byte)0x35,(byte)0xFC,(byte)0x75,(byte)0x0F,(byte)0x92,(byte)0x6E,(byte)0x99,(byte)0x7F,(byte)0xAC,(byte)0x43,(byte)0xAA,(byte)0xE7,(byte)0xD3,(byte)0xA0,(byte)0x3C,(byte)0x81,(byte)0x9B,(byte)0x3C,(byte)0xB8,(byte)0x66,(byte)0xDC,(byte)0xAE,(byte)0xDC,(byte)0xEA,(byte)0x89,(byte)0x2B,(byte)0x00,(byte)0x39,(byte)0xD4,(byte)0x22,(byte)0x7D,(byte)0x01,(byte)0xA0,(byte)0x80,(byte)0xC4,(byte)0x0D,(byte)0x1F,(byte)0x26,(byte)0xF3,(byte)0x18,(byte)0x57,(byte)0x87,(byte)0x0E,(byte)0x3D,(byte)0x49,(byte)0x97,(byte)0x35,(byte)0x39,(byte)0x04,(byte)0xBB,(byte)0xF6,(byte)0xE7,(byte)0xEF,(byte)0xAA,(byte)0xC8,(byte)0x93,(byte)0xEF,(byte)0xCB,(byte)0xA4,(byte)0xAE,(byte)0xB2,(byte)0xF0,(byte)0x55,(byte)0x4A,(byte)0x0F,(byte)0xFE,(byte)0x1A,(byte)0x10,(byte)0xDD,(byte)0x4F,(byte)0xEA,(byte)0x11,(byte)0x3B,(byte)0xB1,(byte)0x4E,(byte)0x61,(byte)0xC4,(byte)0x99}, (short)0, (short)128);
		issuerPublicKey.setExponent(new byte[]{(byte)0x01,(byte)0x00,(byte)0x01}, (short)0, (short)3);
		
		rsaCipher = Cipher.getInstance(Cipher.ALG_RSA_PKCS1, false);
		desCipher = Cipher.getInstance(Cipher.ALG_DES_CBC_ISO9797_M2, false);
		desMAC = Signature.getInstance(Signature.ALG_DES_MAC8_ISO9797_M2, false);
		
		pin = new OwnerPIN((byte)3, (byte)4);
		pin.update(new byte[]{'1', '2', '3', '4'}, (short)0, (byte)4);
		
		random = RandomData.getInstance(RandomData.ALG_SECURE_RANDOM);
	}
	
	public static void install(byte[] bArray, short bOffset, byte bLength) {
		// GP-compliant JavaCard applet registration
		new radboudcard.RadboudCard().register(bArray, (short) (bOffset + 1),
				bArray[bOffset]);
	}

	public void process(APDU apdu) {
		// Good practice: Return 9000 on SELECT
		if (selectingApplet()) {
			reset();
			return;
		}
		
		byte[] buf = apdu.getBuffer();

		switch (buf[ISO7816.OFFSET_INS]) {
		case Constants.INS_DATA:
			apdu.setIncomingAndReceive();
			
			if(buf[ISO7816.OFFSET_CDATA] == Constants.COMMAND_START) {
				// Set identity of the card
				tmp[0] = 0x00;
				tmp[1] = 0x01;
				 
				// Generate nonce
				random.generateData(ssc, (short)0, (short)8);
				Util.arrayCopyNonAtomic(ssc, (short)0, tmp, (short)2, (short)8);

				buf[0] = Constants.RESP_SEND;
				rsaCipher.init(issuerPublicKey, Cipher.MODE_ENCRYPT);
				short len = rsaCipher.doFinal(tmp, (short)0, (short)10, buf, (short)1);

				state[0] = Constants.STATE_INIT;
				
				apdu.setOutgoingAndSend(_0, (short)(len + 1));
			}			
			else if(state[0] == Constants.STATE_INIT && buf[ISO7816.OFFSET_CDATA] == Constants.COMMAND_AUTH) {
				rsaCipher.init(cardPrivateKey, Cipher.MODE_DECRYPT);
				short len = rsaCipher.doFinal(buf, (short)(ISO7816.OFFSET_CDATA + 1), (short)((short)(0xFF & buf[ISO7816.OFFSET_LC]) - 1), tmp, (short)0);

				// Check for correct length of two nonces
				if(len != 16) {
					reset();
					ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
				}
				
				// Check whether correct nonce was returned
				if(Util.arrayCompare(ssc, (short)0, tmp, (short)0, (short)8) != 0) {
					reset();
					ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
				}
				
				// Copy issuer's nonce to ssc
				Util.arrayCopyNonAtomic(tmp, (short)8, ssc, (short)8, (short)8);
				
				// Generate session key
				random.generateData(tmp, (short)0, (short)16);
				sessionKey.setKey(tmp, (short)0);
				
				// Construct response
				buf[0] = Constants.RESP_SEND;
				
				desCipher.init(sharedKey, Cipher.MODE_ENCRYPT);
				len = desCipher.doFinal(tmp, (short)0, (short)16, buf, (short)1);
				
				// Clear key
				Util.arrayFillNonAtomic(tmp, (short)0, (short)16, (byte)0x00);
				
				// Set new state
				state[0] = Constants.STATE_AUTH;
				state[1] = INT_STATE_INIT;
				
				apdu.setOutgoingAndSend(_0, (short)(len + 1));
			}
			else if(state[0] == Constants.STATE_AUTH) {
				short len = unwrapData(buf, ISO7816.OFFSET_CDATA, (short)(buf[ISO7816.OFFSET_LC] & 0xFF), tmp, _0);
				
				if(state[1] == INT_STATE_INIT) {
					buf[0] = Constants.RESP_DISPLAY;
					Util.arrayCopyNonAtomic(tmp, _0, buf, (short)1, len);
					
					state[1] = INT_STATE_DISPLAY;
					
					apdu.setOutgoingAndSend(_0, (short)(len + 1));
				}
				else {
					state[0] = 0;
					ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
				}
			}
			else {
				state[0] = 0;
				ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
			}
			
			break;

		case Constants.INS_INPUT:
			apdu.setIncomingAndReceive();
			
			if(state[0] == Constants.STATE_AUTH && state[1] == INT_STATE_DISPLAY && buf[ISO7816.OFFSET_CDATA] == Constants.OK) {
				// Ask for PIN
				buf[0] = Constants.RESP_INPUT;
				buf[1] = 0x01;
				buf[2] = 0x04;
				buf[3] = 0x50;
				buf[4] = 0x49;
				buf[5] = 0x4E;
				buf[6] = 0x3A;
				
				state[1] = INT_STATE_ASK_PIN;
				
				apdu.setOutgoingAndSend(_0, (short)7);				
			}
			else if(state[0] == Constants.STATE_AUTH && state[1] == INT_STATE_ASK_PIN && buf[ISO7816.OFFSET_CDATA] == Constants.OK) {
				if(!pin.check(buf, (short)(ISO7816.OFFSET_CDATA + 1), (byte)4)) {
					state[0] = 0;
					state[1] = 0;
					ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
				}
				
				// Send response to terminal
				tmp[0] = 'O';
				tmp[1] = 'K';
				
				buf[0] = Constants.RESP_SEND;

				short len = wrapData(tmp, _0, (short)2, buf, (short)1);
				
				state[1] = INT_STATE_INIT;
				
				apdu.setOutgoingAndSend(_0, (short)(len + 1));
			}

			else {
				ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED); 
			}

			break;
	
		default:
			// good practice: If you don't know the INStruction, say so:
			ISOException.throwIt(ISO7816.SW_INS_NOT_SUPPORTED);
		}
	}
	
	public short unwrapData(byte[] in, short in_offset, short in_len, byte[] out, short out_offset) {
		// Initialise DES MAC
		desMAC.init(sessionKey, Signature.MODE_VERIFY);
		desMAC.update(ssc, (short)0, (short)16);
		
		// Verify MAC
		if(!desMAC.verify(in, in_offset, (short)(in_len - 8), in, (short)(in_offset + in_len - 8), (short)8)) {
			reset();
			ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
		}
		
		// Initialise DES decryption
		desCipher.init(sessionKey, Cipher.MODE_DECRYPT);
		
		// Decrypt data
		short out_len = desCipher.doFinal(in, in_offset, (short)(in_len - 8), out, out_offset);
		
		// Increment SSC
		incrementSSC();
		
		return out_len;
	}
	
	public short wrapData(byte[] in, short in_offset, short in_len, byte[] out, short out_offset) {
		// Initialise DES encryption
		desCipher.init(sessionKey, Cipher.MODE_ENCRYPT);
		
		// Encrypt data
		short out_len = desCipher.doFinal(in, in_offset, in_len, out, out_offset);
		
		// Initialise DES MAC
		desMAC.init(sessionKey, Signature.MODE_SIGN);
		desMAC.update(ssc, (short)0, (short)16);
		
		// Generate MAC
		out_len += desMAC.sign(out, out_offset, out_len, out, (short)(out_offset + out_len));
		
		// Increment SSC
		incrementSSC();
		
		return out_len;
	}	
	
	public void reset() {
		state[0] = 0;
		state[1] = 0;
		Util.arrayFillNonAtomic(ssc, (short)0, (short)16, (byte)0x00);
	}

    public void incrementSSC() {
        if (ssc == null || ssc.length <= 0)
            return;

        for (short s = (short) (ssc.length - 1); s >= 0; s--) {
            if ((short) ((ssc[s] & 0xff) + 1) > 0xff)
                ssc[s] = 0;
            else {
                ssc[s]++;
                break;
            }
        }
    }	
}